# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from inductiva.client.models.credits_boost_percentage import CreditsBoostPercentage
from inductiva.client.models.org_status import OrgStatus
from typing import Optional, Set
from typing_extensions import Self


class OrganizationPartialUpdate(BaseModel):
    """
    OrganizationPartialUpdate
    """

  # noqa: E501
    identity_provider_id: Optional[StrictStr] = None
    name: Optional[StrictStr] = None
    tier: Optional[StrictStr] = None
    status: Optional[OrgStatus] = None
    creation_ts: Optional[datetime] = None
    credits_boost_percentage: Optional[CreditsBoostPercentage] = None
    __properties: ClassVar[List[str]] = [
        "identity_provider_id", "name", "tier", "status", "creation_ts",
        "credits_boost_percentage"
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of OrganizationPartialUpdate from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of credits_boost_percentage
        if self.credits_boost_percentage:
            _dict[
                'credits_boost_percentage'] = self.credits_boost_percentage.to_dict(
                )
        # set to None if identity_provider_id (nullable) is None
        # and model_fields_set contains the field
        if self.identity_provider_id is None and "identity_provider_id" in self.model_fields_set:
            _dict['identity_provider_id'] = None

        # set to None if name (nullable) is None
        # and model_fields_set contains the field
        if self.name is None and "name" in self.model_fields_set:
            _dict['name'] = None

        # set to None if tier (nullable) is None
        # and model_fields_set contains the field
        if self.tier is None and "tier" in self.model_fields_set:
            _dict['tier'] = None

        # set to None if status (nullable) is None
        # and model_fields_set contains the field
        if self.status is None and "status" in self.model_fields_set:
            _dict['status'] = None

        # set to None if creation_ts (nullable) is None
        # and model_fields_set contains the field
        if self.creation_ts is None and "creation_ts" in self.model_fields_set:
            _dict['creation_ts'] = None

        # set to None if credits_boost_percentage (nullable) is None
        # and model_fields_set contains the field
        if self.credits_boost_percentage is None and "credits_boost_percentage" in self.model_fields_set:
            _dict['credits_boost_percentage'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of OrganizationPartialUpdate from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "identity_provider_id":
                obj.get("identity_provider_id"),
            "name":
                obj.get("name"),
            "tier":
                obj.get("tier"),
            "status":
                obj.get("status"),
            "creation_ts":
                obj.get("creation_ts"),
            "credits_boost_percentage":
                CreditsBoostPercentage.from_dict(
                    obj["credits_boost_percentage"])
                if obj.get("credits_boost_percentage") is not None else None
        })
        return _obj
